# -*- coding:utf-8 -*-

import asyncio
import contextlib


def _aiorun(f):
    with contextlib.closing(asyncio.new_event_loop()) as loop:
        loop.run_until_complete(f)


def _get_all(reader):
    def get_all(routine, json_content, *args, **kwargs):
        events = []
        async def run():
            async for event in routine(reader(json_content), *args, **kwargs):
                events.append(event)
        _aiorun(run())
        return events
    return get_all


def _get_first(reader):
    def get_first(routine, json_content, *args, **kwargs):
        events = []
        async def run():
            async for event in routine(reader(json_content), *args, **kwargs):
                events.append(event)
                if events:
                    return
        _aiorun(run())
        return events[0]
    return get_first



import os
import subprocess
import sys
import unittest

from ijson import compat
from test.test_base import JSON


class DumpTests(unittest.TestCase):

    def _do_test_dump(self, method, multiple_values):
        # Use python backend to ensure multiple_values works
        env = dict(os.environ)
        env['IJSON_BACKEND'] = 'python'
        # Ensure printing works on the subprocess in Windows
        # by using utf-8 on its stdout
        if 'win' in sys.platform:
            env = dict(os.environ)
            env['PYTHONIOENCODING'] = 'utf-8'
        cmd = [sys.executable, '-m', 'ijson.dump', '-m', method, '-p', '']
        if multiple_values:
            cmd.append('-M')
        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
        input_data = JSON
        if multiple_values:
            input_data += JSON
        out, err = proc.communicate(input_data)
        status = proc.wait()
        self.assertEqual(0, status, "out:\n%s\nerr:%s" % (compat.b2s(out), compat.b2s(err)))

    def _test_dump(self, method):
        self._do_test_dump(method, True)
        self._do_test_dump(method, False)

    def test_basic_parse(self):
        self._test_dump('basic_parse')

    def test_parse(self):
        self._test_dump('parse')

    def test_kvitems(self):
        self._test_dump('kvitems')

    def test_items(self):
        self._test_dump('items')


from ijson import common, compat

from .test_base import (JSON, FileBasedTests, JSON_EVENTS,
    PARTIAL_ARRAY_JSONS, warning_catcher, INVALID_JSONS, IS_PY2,
    generate_test_cases)



class SingleReadFile(object):
    '''A bytes file that can be read only once'''

    def __init__(self, raw_value):
        self.raw_value = raw_value

    def read(self, size=-1):
        if size == 0:
            return compat.bytetype()
        val = self.raw_value
        if not val:
            raise AssertionError('read twice')
        self.raw_value = compat.bytetype()
        return val


class GeneratorSpecificTests(FileBasedTests):
    '''
    Base class for parsing tests that is used to create test cases for each
    available backends.
    '''

    def test_utf8_split(self):
        buf_size = JSON.index(b'\xd1') + 1
        try:
            self.get_all(self.basic_parse, JSON, buf_size=buf_size)
        except UnicodeDecodeError:
            self.fail('UnicodeDecodeError raised')

    def test_lazy(self):
        # shouldn't fail since iterator is not exhausted
        self.backend.basic_parse(compat.BytesIO(INVALID_JSONS[0]))
        self.assertTrue(True)

    def test_boundary_lexeme(self):
        buf_size = JSON.index(b'false') + 1
        events = self.get_all(self.basic_parse, JSON, buf_size=buf_size)
        self.assertEqual(events, JSON_EVENTS)

    def test_boundary_whitespace(self):
        buf_size = JSON.index(b'   ') + 1
        events = self.get_all(self.basic_parse, JSON, buf_size=buf_size)
        self.assertEqual(events, JSON_EVENTS)

    def test_item_building_greediness(self):
        self._test_item_iteration_validity(compat.BytesIO)

    def test_lazy_file_reading(self):
        if self.backend_name == 'python' and IS_PY2:
            # We know it doesn't work because because the decoder itself
            # is quite eager on its reading
            return
        self._test_item_iteration_validity(SingleReadFile)

    def _test_item_iteration_validity(self, file_type):
        for json in PARTIAL_ARRAY_JSONS:
            json, expected_items = json[0], json[1:]
            iterable = self.backend.items(file_type(json), 'item')
            for expected_item in expected_items:
                self.assertEqual(expected_item, next(iterable))


    COMMON_DATA = b'''
        {
            "skip": "skip_value",
            "c": {"d": "e", "f": "g"},
            "list": [{"o1": 1}, {"o2": 2}]
        }'''

    COMMON_PARSE = [
        ('', 'start_map', None),
        ('', 'map_key', 'skip'),
        ('skip', 'string', 'skip_value'),
        ('', 'map_key', 'c'),
        ('c', 'start_map', None),
        ('c', 'map_key', 'd'),
        ('c.d', 'string', 'e'),
        ('c', 'map_key', 'f'),
        ('c.f', 'string', 'g'),
        ('c', 'end_map', None),
        ('', 'map_key', 'list'),
        ('list', 'start_array', None),
        ('list.item', 'start_map', None),
        ('list.item', 'map_key', 'o1'),
        ('list.item.o1', 'number', 1),
        ('list.item', 'end_map', None),
        ('list.item', 'start_map', None),
        ('list.item', 'map_key', 'o2'),
        ('list.item.o2', 'number', 2),
        ('list.item', 'end_map', None),
        ('list', 'end_array', None),
        ('', 'end_map', None),
    ]

    def _skip_parse_events(self, events):
        skip_value = None
        for prefix, _, value in events:
            if prefix == 'skip':
                skip_value = value
                break
        self.assertEqual(skip_value, 'skip_value')

    def _test_common_routine(self, routine, *args, **kwargs):
        base_routine_name = kwargs.pop('base_routine_name', 'parse')
        base_routine = getattr(self.backend, base_routine_name)
        events = base_routine(compat.BytesIO(self.COMMON_DATA))
        if base_routine_name == 'parse':
            self._skip_parse_events(events)
        # Rest of events can still be used
        return list(routine(events, *args))

    def test_common_parse(self):
        with warning_catcher() as warns:
            results = self._test_common_routine(common.parse,
                                                base_routine_name='basic_parse')
        self.assertEqual(self.COMMON_PARSE, results)
        self.assertEqual(len(warns), 1)

    def test_common_kvitems(self):
        with warning_catcher() as warns:
            results = self._test_common_routine(common.kvitems, 'c')
        self.assertEqual([("d", "e"), ("f", "g")], results)
        self.assertEqual(len(warns), 1)

    def test_common_items(self):
        with warning_catcher() as warns:
            results = self._test_common_routine(common.items, 'list.item')
        self.assertEqual([{"o1": 1}, {"o2": 2}], results)
        self.assertEqual(len(warns), 1)


def _reader(json):
    if type(json) == compat.bytetype:
        return compat.BytesIO(json)
    return compat.StringIO(json)


def get_all(routine, json_content, *args, **kwargs):
    return list(routine(_reader(json_content), *args, **kwargs))


def get_first(routine, json_content, *args, **kwargs):
    return next(routine(_reader(json_content), *args, **kwargs))


generate_test_cases(globals(), 'Generators', '_gen', GeneratorSpecificTests)


#
# Contributed by Rodrigo Tobar <rtobar@icrar.org>
#
# ICRAR - International Centre for Radio Astronomy Research
# (c) UWA - The University of Western Australia, 2016
# Copyright by UWA (in the framework of the ICRAR)
#
'''
Wrapper for _yajl2 C extension module
'''

from ijson import common, compat, utils
from . import _yajl2


_get_buf_size = lambda kwargs: kwargs.pop('buf_size', 64 * 1024)

@utils.coroutine
def basic_parse_basecoro(target, **kwargs):
    return _yajl2.basic_parse_basecoro(target.send, **kwargs)

def basic_parse_gen(file, **kwargs):
    f = compat.bytes_reader(file)
    buf_size = _get_buf_size(kwargs)
    return _yajl2.basic_parse(f, buf_size, **kwargs)

def basic_parse_async(file, **kwargs):
    buf_size = _get_buf_size(kwargs)
    return _yajl2.basic_parse_async(file, buf_size, **kwargs)

@utils.coroutine
def parse_basecoro(target, **kwargs):
    return _yajl2.parse_basecoro(target.send, **kwargs)

def parse_gen(file, **kwargs):
    f = compat.bytes_reader(file)
    buf_size = _get_buf_size(kwargs)
    return _yajl2.parse(f, buf_size, **kwargs)

def parse_async(file, **kwargs):
    buf_size = _get_buf_size(kwargs)
    return _yajl2.parse_async(file, buf_size, **kwargs)

@utils.coroutine
def kvitems_basecoro(target, prefix, map_type=None, **kwargs):
    return _yajl2.kvitems_basecoro(target.send, prefix, map_type, **kwargs)

def kvitems_gen(file, prefix, map_type=None, **kwargs):
    f = compat.bytes_reader(file)
    buf_size = _get_buf_size(kwargs)
    return _yajl2.kvitems(f, buf_size, prefix, map_type, **kwargs)

def kvitems_async(file, prefix, map_type=None, **kwargs):
    buf_size = _get_buf_size(kwargs)
    return _yajl2.kvitems_async(file, buf_size, prefix, map_type, **kwargs)

@utils.coroutine
def items_basecoro(target, prefix, map_type=None, **kwargs):
    return _yajl2.items_basecoro(target.send, prefix, map_type, **kwargs)

def items_gen(file, prefix, map_type=None, **kwargs):
    f = compat.bytes_reader(file)
    buf_size = _get_buf_size(kwargs)
    return _yajl2.items(f, buf_size, prefix, map_type, **kwargs)

def items_async(file, prefix, map_type=None, **kwargs):
    buf_size = _get_buf_size(kwargs)
    return _yajl2.items_async(file, buf_size, prefix, map_type, **kwargs)

common.enrich_backend(globals())



'''
Iterative JSON parser.

Main API:

- ``ijson.parse``: iterator returning parsing events with the object tree context,
  see ``ijson.common.parse`` for docs.

- ``ijson.items``: iterator returning Python objects found under a specified prefix,
  see ``ijson.common.items`` for docs.

Top-level ``ijson`` module exposes method from the pure Python backend. There's
also two other backends using the C library yajl in ``ijson.backends`` that have
the same API and are faster under CPython.
'''
from ijson.common import JSONError, IncompleteJSONError, ObjectBuilder, compat

from ijson.utils import coroutine, sendable_list
from .version import __version__

def get_backend(backend):
    """Import the backend named ``backend``"""
    import importlib
    return importlib.import_module('ijson.backends.' + backend)

def _default_backend():
    import os
    if 'IJSON_BACKEND' in os.environ:
        return get_backend(os.environ['IJSON_BACKEND'])
    for backend in ('yajl2_c', 'yajl2_cffi', 'yajl2', 'yajl', 'python'):
        try:
            return get_backend(backend)
        except ImportError:
            continue
    raise ImportError('no backends available')
backend = _default_backend()
del _default_backend

basic_parse = backend.basic_parse
basic_parse_coro = backend.basic_parse_coro
parse = backend.parse
parse_coro = backend.parse_coro
items = backend.items
items_coro = backend.items_coro
kvitems = backend.kvitems
kvitems_coro = backend.kvitems_coro
if compat.IS_PY35:
    basic_parse_async = backend.basic_parse_async
    parse_async = backend.parse_async
    items_async = backend.items_async
    kvitems_async = backend.kvitems_async
backend = backend.backend


# -*- coding:utf-8 -*-

import io
import types

from ijson import compat

from ._test_async_common import _get_all, _get_first

class AsyncReaderTypesCoroutine(object):
    def __init__(self, data):
        if type(data) == compat.bytetype:
            self.data = io.BytesIO(data)
        else:
            self.data = io.StringIO(data)

    async def _read(self, n=-1):
        return self.data.read(n)

    @types.coroutine
    def read(self, n=-1):
        return (yield from self._read(n))

get_all = _get_all(AsyncReaderTypesCoroutine)
get_first = _get_first(AsyncReaderTypesCoroutine)



# -*- coding:utf-8 -*-

import asyncio
import io

from ijson import compat

from ._test_async_common import _get_all, _get_first


class AsyncReader(object):
    def __init__(self, data):
        if type(data) == compat.bytetype:
            self.data = io.BytesIO(data)
        else:
            self.data = io.StringIO(data)

    async def read(self, n=-1):
        await asyncio.sleep(0)
        return self.data.read(n)

get_all = _get_all(AsyncReader)
get_first = _get_first(AsyncReader)



#
# Contributed by Rodrigo Tobar <rtobar@icrar.org>
#
# ICRAR - International Centre for Radio Astronomy Research
# (c) UWA - The University of Western Australia, 2019
# Copyright by UWA (in the framework of the ICRAR)
#
'''
Benchmarking utility for ijson
'''
import argparse
import collections
import io
import os
import sys
import time


import ijson
from ijson import compat

_benchmarks = collections.OrderedDict()
def benchmark(f):
    _benchmarks[f.__name__] = f
    return f

@benchmark
def long_list(n):
    return b'[' + b','.join([b'1' for _ in range(n)]) + b']'

@benchmark
def big_int_object(n):
    return b'{' + b',\n'.join([b'"key_%d": %d' % (i, i) for i in range(n)]) + b'}'

@benchmark
def big_decimal_object(n):
    return b'{' + b',\n'.join([b'"key_%d": %d.0' % (i, i) for i in range(n)]) + b'}'

@benchmark
def big_null_object(n):
    return b'{' + b',\n'.join([b'"key_%d": null' % (i,) for i in range(n)]) + b'}'

@benchmark
def big_bool_object(n):
    return b'{' + b',\n'.join([
        b'"key_%d": %s' % (i, b"true" if i % 2 == 0 else b"false")
        for i in range(n)]) + b'}'

@benchmark
def big_str_object(n):
    return b'{' + b',\n'.join([b'"key_%d": "value_%d"' % (i, i) for i in range(n)]) + b'}'

@benchmark
def big_longstr_object(n):
    str_template = b"value that is very long and should cause a bit less of JSON parsing"
    return b'{' + b',\n'.join([b'"key_%d": "%s"' % (i, str_template) for i in range(n)]) + b'}'

@benchmark
def object_with_10_keys(n):
    template = b'{' + b',\n'.join([b'"key_%d": "value_%d"' % (i, i) for i in range(10)]) + b'}'
    return b'[' + b',\n'.join(
        template
        for _ in range(n)) + b']'

@benchmark
def empty_lists(n):
    return b'[' + b', '.join(b'[]' for _ in range(n)) + b']'

@benchmark
def empty_objects(n):
    return b'[' + b', '.join(b'{}' for _ in range(n)) + b']'

def parse_benchmarks(s):
    return [_benchmarks[name] for name in s.split(',')]


BACKEND_NAMES = 'python', 'yajl', 'yajl2', 'yajl2_cffi', 'yajl2_c'

def load_backends():
    backends = collections.OrderedDict()
    for backend_name in BACKEND_NAMES:
        try:
            backends[backend_name] = ijson.get_backend(backend_name)
        except ImportError:
            continue
    return backends
_backends = load_backends()

def parse_backends(s):
    backends = collections.OrderedDict()
    for name in s.split(','):
        backends[name] = _backends[name]
    return backends


def _stdout_tty_write_flush(message):
    stdout = sys.stdout
    if stdout.isatty():
        stdout.write(message)
        stdout.flush()


class progress_message(object):

    def __init__(self, message):
        self.message = message

    def __enter__(self):
        _stdout_tty_write_flush(self.message)
        return self

    def __exit__(self, *args):
        _stdout_tty_write_flush('\r\033[K')


if compat.IS_PY35:
    exec('''
class AsyncReader(object):
    def __init__(self, data):
        self.data = io.BytesIO(data)

    async def read(self, n=-1):
        return self.data.read(n)

    def close(self):
        self.data.close()

async def _run_async(method, reader, *method_args, **method_kwargs):
    async for _ in method(reader, *method_args, **method_kwargs):
        pass
    ''')

def run_benchmarks(args, benchmark_func=None, fname=None):
    if bool(benchmark_func) == bool(fname):
        raise ValueError("Either benchmark_func or fname must be given")
    if benchmark_func:
        bname = benchmark_func.__name__
        with progress_message('Generating data for benchmark %s...' % (bname,)):
            data = benchmark_func(args.size)
            size = len(data)
    else:
        bname = fname
        size = os.stat(args.input).st_size

    for backend_name, backend in args.backends.items():

        # Get correct method and prepare its arguments
        method = args.method
        if args.run_async:
            method += '_async'
        elif args.run_coro:
            method += '_coro'
        method = getattr(backend, method)
        method_args = ()
        if args.method in ('items', 'kvitems'):
            method_args = args.prefix,
        method_kwargs = {
            'multiple_values': args.multiple_values,
            'use_float': args.use_float
        }
        if not args.run_coro:
            method_kwargs['buf_size'] = args.bufsize

        # Prepare reader
        reader = None
        if not benchmark_func:
            reader = open(fname, 'rb')
        else:
            reader = AsyncReader(data) if args.run_async else io.BytesIO(data)

        # Prepare function that will run the benchmark
        if args.run_async:
            import asyncio
            loop = asyncio.new_event_loop()
            def run():
                try:
                    loop.run_until_complete(_run_async(method, reader, *method_args, **method_kwargs))
                finally:
                    loop.close()
        elif args.run_coro:
            def run():
                from ijson.utils import sendable_list
                events = sendable_list()
                coro = method(events, *method_args, **method_kwargs)
                if reader:
                    chunks = iter(lambda: reader.read(args.bufsize), b'')
                else:
                    chunks = (data[pos:pos + args.bufsize]
                              for pos in range(0, len(data), args.bufsize))
                for chunk in chunks:
                    coro.send(chunk)
                    del events[:]
                coro.close()
        else:
            def run():
                for _ in method(reader, *method_args, **method_kwargs):
                    pass

        # Go, go, go!
        start = time.time()
        run()
        duration = time.time() - start
        megabytes = size / 1024. / 1024.
        print("%.3f, %s, %s, %s, %.3f, %.3f" %
              (megabytes, args.method, bname, backend_name, duration,
               megabytes / duration))
        reader.close()


def main():
    DEFAULT_N = 100000
    DEFAULT_BUFSIZE = 64 * 1024
    ALL_BENCHMARKS = ','.join(_benchmarks)
    ALL_BACKENDS = ','.join(_backends)
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--size', type=int,
        help='Size of JSON content; actual size in bytes might differ, defaults to %d' % DEFAULT_N,
        default=DEFAULT_N)
    parser.add_argument('-S', '--bufsize', type=int,
        help='Buffer size used during parsing; defaults to %d' % DEFAULT_BUFSIZE,
        default=DEFAULT_BUFSIZE)
    parser.add_argument('-b', '--benchmarks', type=parse_benchmarks,
        help='Comma-separated list of benchmarks to include, defaults to %s' % ALL_BENCHMARKS,
        default=ALL_BENCHMARKS)
    parser.add_argument('-B', '--backends', type=parse_backends,
        help='Comma-separated list of backends to include, defaults to %s' % ALL_BACKENDS,
        default=ALL_BACKENDS)
    parser.add_argument('-l', '--list', action='store_true',
        help='List available benchmarks and backends')
    parser.add_argument('-i', '--input',
        help='File to use for benchmarks rather than built-in benchmarking functions')
    parser.add_argument('-m', '--multiple-values', action='store_true', default=False,
        help='Content has multiple JSON values, useful when used with -i')
    parser.add_argument('-f', '--use-float', action='store_true', default=False,
        help='Parse non-integer numbers as float instead of Decimal')
    parser.add_argument('-M', '--method', choices=['basic_parse', 'parse', 'kvitems', 'items'],
                        help='The method to benchmark', default='basic_parse')
    parser.add_argument('-c', '--coro', action='store_true', default=False,
                        dest='run_coro', help='Benchmark coroutine methods')
    if compat.IS_PY35:
        parser.add_argument('-a', '--async', action='store_true', default=False,
                            dest='run_async', help='Benchmark asyncio-enabled methods')
    parser.add_argument('-p', '--prefix', help='Prefix (used with -M items|kvitems)', default='')

    args = parser.parse_args()
    if args.list:
        msg = 'Backends:\n'
        msg += '\n'.join(' - %s' % name for name in _backends)
        msg += '\nBenchmarks:\n'
        msg += '\n'.join(' - %s' % name for name in _benchmarks)
        print(msg)
        return

    print("#mbytes,method,test_case,backend,time,mb_per_sec")
    if args.input:
        run_benchmarks(args, fname=args.input)
    else:
        for benchmark in args.benchmarks:
            run_benchmarks(args, benchmark)

if __name__ == '__main__':
    main()



from ijson import compat

from .test_base import FileBasedTests, generate_test_cases


# Generating real TestCase classes for each importable backend
if compat.IS_PY35:
    from ._test_async import *  # @UnusedWildImport
    generate_test_cases(globals(), 'Async', '_async', FileBasedTests)


'''
Wrapper for YAJL C library version 2.x.
'''

from ctypes import byref

from ijson import common, utils
from ijson.backends import _yajl2_ctypes_common


yajl = _yajl2_ctypes_common.get_yajl(2)

# constants defined in yajl_parse.h
YAJL_ALLOW_COMMENTS = 1
YAJL_MULTIPLE_VALUES = 8


@utils.coroutine
def basic_parse_basecoro(target, allow_comments=False, multiple_values=False,
                         use_float=False):
    '''
    Iterator yielding unprefixed events.

    Parameters:

    - f: a readable file-like object with JSON input
    - allow_comments: tells parser to allow comments in JSON input
    - buf_size: a size of an input buffer
    - multiple_values: allows the parser to parse multiple JSON objects
    '''
    callbacks = _yajl2_ctypes_common.make_callbaks(target.send, use_float, 2)
    handle = yajl.yajl_alloc(byref(callbacks), None, None)
    if allow_comments:
        yajl.yajl_config(handle, YAJL_ALLOW_COMMENTS, 1)
    if multiple_values:
        yajl.yajl_config(handle, YAJL_MULTIPLE_VALUES, 1)
    try:
        while True:
            try:
                buffer = (yield)
            except GeneratorExit:
                buffer = b''
            if buffer:
                result = yajl.yajl_parse(handle, buffer, len(buffer))
            else:
                result = yajl.yajl_complete_parse(handle)
            if result != _yajl2_ctypes_common.YAJL_OK:
                error = _yajl2_ctypes_common.yajl_get_error(yajl, handle, buffer)
                exception = common.IncompleteJSONError if result == _yajl2_ctypes_common.YAJL_INSUFFICIENT_DATA else common.JSONError
                raise exception(error)
            if not buffer:
                break
    finally:
        yajl.yajl_free(handle)


common.enrich_backend(globals())



from ijson import utils, compat

from .test_base import generate_test_cases


if compat.IS_PY2:
    def bytesiter(x):
        return x
else:
    def bytesiter(x):
        for b in x:
            yield bytes([b])


def get_all(routine, json_content, *args, **kwargs):
    events = utils.sendable_list()
    coro = routine(events, *args, **kwargs)
    for datum in bytesiter(json_content):
        coro.send(datum)
    coro.close()
    return events


def get_first(routine, json_content, *args, **kwargs):
    events = utils.sendable_list()
    coro = routine(events, *args, **kwargs)
    for datum in bytesiter(json_content):
        coro.send(datum)
        if events:
            return events[0]
    coro.close()
    if events:
        return events[0]
    return None

generate_test_cases(globals(), 'Coroutines', '_coro')


# -*- coding:utf-8 -*-
from __future__ import unicode_literals
import collections
import ctypes
import unittest
from decimal import Decimal
import threading

import ijson
from ijson import common, compat
from ijson.compat import b2s, IS_PY2
import warnings


JSON = b'''
{
  "docs": [
    {
      "null": null,
      "boolean": false,
      "true": true,
      "integer": 0,
      "double": 0.5,
      "exponent": 1.0e+2,
      "long": 10000000000,
      "string": "\\u0441\\u0442\\u0440\\u043e\\u043a\\u0430 - \xd1\x82\xd0\xb5\xd1\x81\xd1\x82",
      "\xc3\xb1and\xc3\xba": null
    },
    {
      "meta": [[1], {}]
    },
    {
      "meta": {"key": "value"}
    },
    {
      "meta": null
    },
    {
      "meta": []
    }
  ]
}
'''
JSON_OBJECT = {
    "docs": [
        {
            "null": None,
            "boolean": False,
            "true": True,
            "integer": 0,
            "double": Decimal("0.5"),
            "exponent": 1e+2,
            "long": 10000000000,
            "string": "строка - тест",
            "ñandú": None
        },
        {
            "meta": [[1], {}]
        },
        {
            "meta": {
                "key": "value"
            }
        },
        {
            "meta": None
        },
        {
            "meta": []
        }
    ]
}
JSON_PARSE_EVENTS = [
    ('', 'start_map', None),
    ('', 'map_key', 'docs'),
    ('docs', 'start_array', None),
    ('docs.item', 'start_map', None),
    ('docs.item', 'map_key', 'null'),
    ('docs.item.null', 'null', None),
    ('docs.item', 'map_key', 'boolean'),
    ('docs.item.boolean', 'boolean', False),
    ('docs.item', 'map_key', 'true'),
    ('docs.item.true', 'boolean', True),
    ('docs.item', 'map_key', 'integer'),
    ('docs.item.integer', 'number', 0),
    ('docs.item', 'map_key', 'double'),
    ('docs.item.double', 'number', Decimal('0.5')),
    ('docs.item', 'map_key', 'exponent'),
    ('docs.item.exponent', 'number', Decimal('1.0E+2')),
    ('docs.item', 'map_key', 'long'),
    ('docs.item.long', 'number', 10000000000),
    ('docs.item', 'map_key', 'string'),
    ('docs.item.string', 'string', 'строка - тест'),
    ('docs.item', 'map_key', 'ñandú'),
    ('docs.item.ñandú', 'null', None),
    ('docs.item', 'end_map', None),
    ('docs.item', 'start_map', None),
    ('docs.item', 'map_key', 'meta'),
    ('docs.item.meta', 'start_array', None),
    ('docs.item.meta.item', 'start_array', None),
    ('docs.item.meta.item.item', 'number', 1),
    ('docs.item.meta.item', 'end_array', None),
    ('docs.item.meta.item', 'start_map', None),
    ('docs.item.meta.item', 'end_map', None),
    ('docs.item.meta', 'end_array', None),
    ('docs.item', 'end_map', None),
    ('docs.item', 'start_map', None),
    ('docs.item', 'map_key', 'meta'),
    ('docs.item.meta', 'start_map', None),
    ('docs.item.meta', 'map_key', 'key'),
    ('docs.item.meta.key', 'string', 'value'),
    ('docs.item.meta', 'end_map', None),
    ('docs.item', 'end_map', None),
    ('docs.item', 'start_map', None),
    ('docs.item', 'map_key', 'meta'),
    ('docs.item.meta', 'null', None),
    ('docs.item', 'end_map', None),
    ('docs.item', 'start_map', None),
    ('docs.item', 'map_key', 'meta'),
    ('docs.item.meta', 'start_array', None),
    ('docs.item.meta', 'end_array', None),
    ('docs.item', 'end_map', None),
    ('docs', 'end_array', None),
    ('', 'end_map', None)
]
JSON_KVITEMS = [
    ("null", None),
    ("boolean", False),
    ("true", True),
    ("integer", 0),
    ("double", Decimal("0.5")),
    ("exponent", 1e+2),
    ("long", 10000000000),
    ("string", "строка - тест"),
    ("ñandú", None),
    ("meta", [[1], {}]),
    ("meta", {"key": "value"}),
    ("meta", None),
    ("meta", [])
]
JSON_KVITEMS_META = [
    ('key', 'value')
]
JSON_EVENTS = [
    ('start_map', None),
        ('map_key', 'docs'),
        ('start_array', None),
            ('start_map', None),
                ('map_key', 'null'),
                ('null', None),
                ('map_key', 'boolean'),
                ('boolean', False),
                ('map_key', 'true'),
                ('boolean', True),
                ('map_key', 'integer'),
                ('number', 0),
                ('map_key', 'double'),
                ('number', Decimal('0.5')),
                ('map_key', 'exponent'),
                ('number', 100),
                ('map_key', 'long'),
                ('number', 10000000000),
                ('map_key', 'string'),
                ('string', 'строка - тест'),
                ('map_key', 'ñandú'),
                ('null', None),
            ('end_map', None),
            ('start_map', None),
                ('map_key', 'meta'),
                ('start_array', None),
                    ('start_array', None),
                        ('number', 1),
                    ('end_array', None),
                    ('start_map', None),
                    ('end_map', None),
                ('end_array', None),
            ('end_map', None),
            ('start_map', None),
                ('map_key', 'meta'),
                ('start_map', None),
                    ('map_key', 'key'),
                    ('string', 'value'),
                ('end_map', None),
            ('end_map', None),
            ('start_map', None),
                ('map_key', 'meta'),
                ('null', None),
            ('end_map', None),
            ('start_map', None),
                ('map_key', 'meta'),
                ('start_array', None),
                ('end_array', None),
            ('end_map', None),
        ('end_array', None),
    ('end_map', None),
]

# Like JSON, but with an additional top-level array structure
ARRAY_JSON = b'[' + JSON + b']'
ARRAY_JSON_EVENTS = (
    [('start_array', None)] +
    JSON_EVENTS +
    [('end_array', None)]
)
ARRAY_JSON_PARSE_EVENTS = (
    [('', 'start_array', None)] +
    [('.'.join(filter(None, ('item', p))), t, e) for p, t, e in JSON_PARSE_EVENTS] +
    [('', 'end_array', None)]
)
ARRAY_JSON_OBJECT = [JSON_OBJECT]



SCALAR_JSON = b'0'
INVALID_JSONS = [
    b'["key", "value",]',      # trailing comma
    b'["key"  "value"]',       # no comma
    b'{"key": "value",}',      # trailing comma
    b'{"key": "value" "key"}', # no comma
    b'{"key"  "value"}',       # no colon
    b'invalid',                # unknown lexeme
    b'[1, 2] dangling junk',   # dangling junk
    b'}',                      # no corresponding opening token
    b']',                      # no corresponding opening token
    b'"\xa8"'                  # invalid UTF-8 byte sequence
]
YAJL1_PASSING_INVALID = INVALID_JSONS[6]
INCOMPLETE_JSONS = [
    b'',
    b'"test',
    b'[',
    b'[1',
    b'[1,',
    b'{',
    b'{"key"',
    b'{"key":',
    b'{"key": "value"',
    b'{"key": "value",',
]
INCOMPLETE_JSON_TOKENS = [
    b'n',
    b'nu',
    b'nul',
    b't',
    b'tr',
    b'tru',
    b'f',
    b'fa',
    b'fal',
    b'fals',
    b'[f',
    b'[fa',
    b'[fal',
    b'[fals',
    b'[t',
    b'[tr',
    b'[tru',
    b'[n',
    b'[nu',
    b'[nul',
    b'{"key": t',
    b'{"key": tr',
    b'{"key": tru',
    b'{"key": f',
    b'{"key": fa',
    b'{"key": fal',
    b'{"key": fals',
    b'{"key": n',
    b'{"key": nu',
    b'{"key": nul',
]
STRINGS_JSON = br'''
{
    "str1": "",
    "str2": "\"",
    "str3": "\\",
    "str4": "\\\\",
    "special\t": "\b\f\n\r\t"
}
'''
SURROGATE_PAIRS_JSON = br'"\uD83D\uDCA9"'
PARTIAL_ARRAY_JSONS = [
    (b'[1,', 1),
    (b'[1, 2 ', 1, 2),
    (b'[1, "abc"', 1, 'abc'),
    (b'[{"abc": [0, 1]}', {'abc': [0, 1]}),
    (b'[{"abc": [0, 1]},', {'abc': [0, 1]}),
]

items_test_case = collections.namedtuple('items_test_case', 'json, prefix, kvitems, items')
EMPTY_MEMBER_TEST_CASES = {
    'simple': items_test_case(
        b'{"a": {"": {"b": 1, "c": 2}}}',
        'a.',
        [("b", 1), ("c", 2)],
        [{"b": 1, "c": 2}]
    ),
    'embedded': items_test_case(
        b'{"a": {"": {"": {"b": 1, "c": 2}}}}',
        'a..',
        [("b", 1), ("c", 2)],
        [{"b": 1, "c": 2}]
    ),
    'top_level': items_test_case(
        b'{"": 1, "a": 2}',
        '',
        [("", 1), ("a", 2)],
        [{"": 1, "a": 2}]
    ),
    'top_level_embedded': items_test_case(
        b'{"": {"": 1}, "a": 2}',
        '',
        [("", {"": 1}), ("a", 2)],
        [{"": {"": 1}, "a": 2}]
    )
}


class warning_catcher(object):
    '''Encapsulates proper warning catch-all logic in python 2.7 and 3'''

    def __init__(self):
        self.catcher = warnings.catch_warnings(record=True)

    def __enter__(self):
        ret = self.catcher.__enter__()
        if compat.IS_PY2:
            warnings.simplefilter("always")
        return ret

    def __exit__(self, *args):
        self.catcher.__exit__(*args)


class BackendSpecificTestCase(object):
    '''
    Base class for backend-specific tests, gives ability to easily and
    generically reference different methods on the backend. It requires
    subclasses to define a `backend` member with the backend module, and a
    `suffix` attribute indicating the method flavour to obtain.
    '''

    def __getattr__(self, name):
        return getattr(self.backend, name + self.method_suffix)


class IJsonTestsBase(object):
    '''
    Base class with common tests for all iteration methods.
    Subclasses implement `all()` and `first()` to collect events coming from
    a particuliar method.
    '''

    def test_basic_parse(self):
        events = self.get_all(self.basic_parse, JSON)
        self.assertEqual(events, JSON_EVENTS)

    def test_basic_parse_threaded(self):
        thread = threading.Thread(target=self.test_basic_parse)
        thread.start()
        thread.join()

    def test_parse(self):
        events = self.get_all(self.parse, JSON)
        self.assertEqual(events, JSON_PARSE_EVENTS)

    def test_items(self):
        events = self.get_all(self.items, JSON, '')
        self.assertEqual(events, [JSON_OBJECT])

    def test_items_twodictlevels(self):
        json = b'{"meta":{"view":{"columns":[{"id": -1}, {"id": -2}]}}}'
        ids = self.get_all(self.items, json, 'meta.view.columns.item.id')
        self.assertEqual(2, len(ids))
        self.assertListEqual([-2,-1], sorted(ids))

    def test_items_with_dotted_name(self):
        json = b'{"0.1": 0}'
        self.assertListEqual([0], self.get_all(self.items, json, '0.1'))
        json = b'{"0.1": [{"a.b": 0}]}'
        self.assertListEqual([0], self.get_all(self.items, json, '0.1.item.a.b'))
        json = b'{"0.1": 0, "0": {"1": 1}}'
        self.assertListEqual([0, 1], self.get_all(self.items, json, '0.1'))
        json = b'{"abc.def": 0}'
        self.assertListEqual([0], self.get_all(self.items, json, 'abc.def'))
        self.assertListEqual([], self.get_all(self.items, json, 'abc'))
        self.assertListEqual([], self.get_all(self.items, json, 'def'))

    def test_map_type(self):
        obj = self.get_first(self.items, JSON, '')
        self.assertTrue(isinstance(obj, dict))
        obj = self.get_first(self.items, JSON, '', map_type=collections.OrderedDict)
        self.assertTrue(isinstance(obj, collections.OrderedDict))

    def test_kvitems(self):
        kvitems = self.get_all(self.kvitems, JSON, 'docs.item')
        self.assertEqual(JSON_KVITEMS, kvitems)

    def test_kvitems_toplevel(self):
        kvitems = self.get_all(self.kvitems, JSON, '')
        self.assertEqual(1, len(kvitems))
        key, value = kvitems[0]
        self.assertEqual('docs', key)
        self.assertEqual(JSON_OBJECT['docs'], value)

    def test_kvitems_empty(self):
        kvitems = self.get_all(self.kvitems, JSON, 'docs')
        self.assertEqual([], kvitems)

    def test_kvitems_twodictlevels(self):
        json = b'{"meta":{"view":{"columns":[{"id": -1}, {"id": -2}]}}}'
        view = self.get_all(self.kvitems, json, 'meta.view')
        self.assertEqual(1, len(view))
        key, value = view[0]
        self.assertEqual('columns', key)
        self.assertEqual([{'id': -1}, {'id': -2}], value)

    def test_kvitems_different_underlying_types(self):
        kvitems = self.get_all(self.kvitems, JSON, 'docs.item.meta')
        self.assertEqual(JSON_KVITEMS_META, kvitems)

    def test_basic_parse_array(self):
        events = self.get_all(self.basic_parse, ARRAY_JSON)
        self.assertEqual(events, ARRAY_JSON_EVENTS)

    def test_basic_parse_array_threaded(self):
        thread = threading.Thread(target=self.test_basic_parse_array)
        thread.start()
        thread.join()

    def test_parse_array(self):
        events = self.get_all(self.parse, ARRAY_JSON)
        self.assertEqual(events, ARRAY_JSON_PARSE_EVENTS)

    def test_items_array(self):
        events = self.get_all(self.items, ARRAY_JSON, '')
        self.assertEqual(events, [ARRAY_JSON_OBJECT])

    def test_kvitems_array(self):
        kvitems = self.get_all(self.kvitems, ARRAY_JSON, 'item.docs.item')
        self.assertEqual(JSON_KVITEMS, kvitems)

    def test_scalar(self):
        events = self.get_all(self.basic_parse, SCALAR_JSON)
        self.assertEqual(events, [('number', 0)])

    def test_strings(self):
        events = self.get_all(self.basic_parse, STRINGS_JSON)
        strings = [value for event, value in events if event == 'string']
        self.assertEqual(strings, ['', '"', '\\', '\\\\', '\b\f\n\r\t'])
        self.assertTrue(('map_key', 'special\t') in events)

    def test_surrogate_pairs(self):
        event = self.get_first(self.basic_parse, SURROGATE_PAIRS_JSON)
        parsed_string = event[1]
        self.assertEqual(parsed_string, '💩')

    def test_numbers(self):
        """Check that numbers are correctly parsed"""

        def get_numbers(json, **kwargs):
            events = self.get_all(self.basic_parse, json, **kwargs)
            return events, [value for event, value in events if event == 'number']

        def assert_numbers(json, expected_float_type, *numbers, **kwargs):
            events, values = get_numbers(json, **kwargs)
            float_types = set(type(value) for event, value in events if event == 'number')
            float_types -= {int}
            self.assertEqual(1, len(float_types))
            self.assertEqual(next(iter(float_types)), expected_float_type)
            self.assertSequenceEqual(numbers, values)

        NUMBERS_JSON = b'[1, 1.0, 1E2]'
        assert_numbers(NUMBERS_JSON, Decimal, 1, Decimal("1.0"), Decimal("1e2"))
        assert_numbers(NUMBERS_JSON, float, 1, 1., 100., use_float=True)
        assert_numbers(b'1e400', Decimal, Decimal('1e400'))
        assert_numbers(b'1e-400', Decimal, Decimal('1e-400'))
        assert_numbers(b'1e-400', float, 0., use_float=True)
        # Test for 64-bit integers support when using use_float=True
        try:
            past32bits = 2 ** 32 + 1
            received = get_numbers(('%d' % past32bits).encode('utf8'), use_float=True)[1][0]
            self.assertTrue(self.supports_64bit_integers)
            self.assertEqual(past32bits, received)
        except common.JSONError:
            self.assertFalse(self.supports_64bit_integers)
        # Check that numbers bigger than MAX_DOUBLE cannot be represented
        try:
            get_numbers(b'1e400', use_float=True)
            self.fail("Overflow error expected")
        except common.JSONError:
            pass

    def test_invalid_numbers(self):
        # leading zeros
        if self.detects_leading_zeros:
            for case in (b'00',   b'01',   b'001'):
                for base in (case, case + b'.0', case + b'e0', case + b'E0'):
                    for n in (base, b'-' + base):
                        with self.assertRaises(common.JSONError):
                            self.get_all(self.basic_parse, n)
        # incomplete exponents
        for n in (b'1e', b'0.1e', b'0E'):
            with self.assertRaises(common.JSONError):
                self.get_all(self.basic_parse, n)
        # incomplete fractions
        for n in (b'1.', b'.1'):
            with self.assertRaises(common.JSONError):
                self.get_all(self.basic_parse, n)

    def test_incomplete(self):
        for json in INCOMPLETE_JSONS:
            with self.assertRaises(common.IncompleteJSONError):
                self.get_all(self.basic_parse, json)

    def test_incomplete_tokens(self):
        if not self.handles_incomplete_json_tokens:
            return
        for json in INCOMPLETE_JSON_TOKENS:
            with self.assertRaises(common.IncompleteJSONError):
                self.get_all(self.basic_parse, json)

    def test_invalid(self):
        for json in INVALID_JSONS:
            # Yajl1 doesn't complain about additional data after the end
            # of a parsed object. Skipping this test.
            if self.backend_name == 'yajl' and json == YAJL1_PASSING_INVALID:
                continue
            with self.assertRaises(common.JSONError):
                self.get_all(self.basic_parse, json)

    def test_multiple_values(self):
        """Test that the multiple_values flag works"""
        if not self.supports_multiple_values:
            with self.assertRaises(ValueError):
                self.get_all(self.basic_parse, "", multiple_values=True)
            return
        multiple_json = JSON + JSON + JSON
        items = lambda x, **kwargs: self.items(x, '', **kwargs)
        for func in (self.basic_parse, items):
            with self.assertRaises(common.JSONError):
                self.get_all(func, multiple_json)
            with self.assertRaises(common.JSONError):
                self.get_all(func, multiple_json, multiple_values=False)
            result = self.get_all(func, multiple_json, multiple_values=True)
            if func == items:
                self.assertEqual(result, [JSON_OBJECT, JSON_OBJECT, JSON_OBJECT])
            else:
                self.assertEqual(result, JSON_EVENTS + JSON_EVENTS + JSON_EVENTS)

    def test_comments(self):
        json = b'{"a": 2 /* a comment */}'
        try:
            self.get_all(self.basic_parse, json, allow_comments=True)
        except ValueError:
            if self.supports_comments:
                raise

    def _test_empty_member(self, test_case):
        pairs = self.get_all(self.kvitems, test_case.json, test_case.prefix)
        self.assertEqual(test_case.kvitems, pairs)
        objects = self.get_all(self.items, test_case.json, test_case.prefix)
        self.assertEqual(test_case.items, objects)

    def test_empty_member(self):
        self._test_empty_member(EMPTY_MEMBER_TEST_CASES['simple'])

    def test_embedded_empty_member(self):
        self._test_empty_member(EMPTY_MEMBER_TEST_CASES['embedded'])

    def test_top_level_empty_member(self):
        self._test_empty_member(EMPTY_MEMBER_TEST_CASES['top_level'])

    def test_top_level_embedded_empty_member(self):
        self._test_empty_member(EMPTY_MEMBER_TEST_CASES['top_level_embedded'])


class FileBasedTests(object):

    def test_string_stream(self):
        with warning_catcher() as warns:
            events = self.get_all(self.basic_parse, b2s(JSON))
            self.assertEqual(events, JSON_EVENTS)
        if self.warn_on_string_stream:
            self.assertEqual(len(warns), 1)
            self.assertEqual(DeprecationWarning, warns[0].category)

    def test_different_buf_sizes(self):
        for buf_size in (1, 4, 16, 64, 256, 1024, 4098):
            events = self.get_all(self.basic_parse, JSON, buf_size=buf_size)
            self.assertEqual(events, JSON_EVENTS)


def generate_backend_specific_tests(module, classname_prefix, method_suffix,
                                    *bases, **kwargs):
    for backend in ['python', 'yajl', 'yajl2', 'yajl2_cffi', 'yajl2_c']:
        try:
            classname = '%s%sTests' % (
                ''.join(p.capitalize() for p in backend.split('_')),
                classname_prefix
            )
            if IS_PY2:
                classname = classname.encode('ascii')

            _bases = bases + (BackendSpecificTestCase, unittest.TestCase)
            _members = {
                'backend_name': backend,
                'backend': ijson.get_backend(backend),
                'method_suffix': method_suffix,
                'warn_on_string_stream': not IS_PY2,
                'supports_64bit_integers': not (backend == 'yajl' and ctypes.sizeof(ctypes.c_long) == 4)
            }
            members = kwargs.get('members', lambda _: {})
            _members.update(members(backend))
            module[classname] = type(classname, _bases, _members)
        except ImportError:
            pass


def generate_test_cases(module, classname, method_suffix, *bases):
        _bases = bases + (IJsonTestsBase,)
        members = lambda name: {
            'get_all': lambda self, *args, **kwargs: module['get_all'](*args, **kwargs),
            'get_first': lambda self, *args, **kwargs: module['get_first'](*args, **kwargs),
            'supports_multiple_values': name != 'yajl',
            'supports_comments': name != 'python',
            'detects_leading_zeros': name != 'yajl',
            'handles_incomplete_json_tokens': name != 'yajl'
        }
        return generate_backend_specific_tests(module, classname, method_suffix,
                                               members=members, *_bases)


import importlib.util
import unittest

from ijson import common, compat

from .test_base import warning_catcher
from test.test_base import JSON, JSON_EVENTS, JSON_PARSE_EVENTS, JSON_OBJECT,\
    generate_backend_specific_tests, JSON_KVITEMS


class Misc(unittest.TestCase):
    """Miscellaneous unit tests"""

    def test_common_number_is_deprecated(self):
        with warning_catcher() as warns:
            common.number("1")
        self.assertEqual(len(warns), 1)
        self.assertEqual(DeprecationWarning, warns[0].category)

    def test_yajl2_c_loadable(self):
        if compat.IS_PY2:
            self.skipTest("Test requires Python 3.4+")
        spec = importlib.util.find_spec("ijson.backends._yajl2")
        if spec is None:
            self.skipTest("yajl2_c is not built")
        importlib.util.module_from_spec(spec)


class MainEntryPoints(object):

    def _assert_invalid_type(self, routine, *args, **kwargs):
        # Functions are not valid inputs
        with self.assertRaises(ValueError):
            routine(lambda _: JSON, *args, **kwargs)

    def _assert_bytes(self, expected_results, routine, *args, **kwargs):
        results = list(routine(JSON, *args, **kwargs))
        self.assertEqual(expected_results, results)

    def _assert_str(self, expected_results, routine, *args, **kwargs):
        with warning_catcher() as warns:
            results = list(routine(compat.b2s(JSON), *args, **kwargs))
        self.assertEqual(expected_results, results)
        if self.warn_on_string_stream:
            self.assertEqual(1, len(warns))

    def _assert_unicode(self, expected_results, routine, *args, **kwargs):
        if not compat.IS_PY2:
            return
        with warning_catcher() as warns:
            results = list(routine(unicode(JSON, 'utf-8'), *args, **kwargs))
        self.assertEqual(expected_results, results)
        self.assertEqual(1, len(warns))

    def _assert_file(self, expected_results, routine, *args, **kwargs):
        results = list(routine(compat.BytesIO(JSON), *args, **kwargs))
        self.assertEqual(expected_results, results)

    def _assert_async_file(self, expected_results, routine, *args, **kwargs):
        if not compat.IS_PY35:
            return
        from ._test_async import get_all
        results = get_all(routine, JSON, *args, **kwargs)
        self.assertEqual(expected_results, results)

    def _assert_async_types_coroutine(self, expected_results, routine, *args, **kwargs):
        if not compat.IS_PY35:
            return
        from ._test_async_types_coroutine import get_all
        results = get_all(routine, JSON, *args, **kwargs)
        self.assertEqual(expected_results, results)

    def _assert_events(self, expected_results, previous_routine, routine, *args, **kwargs):
        events = previous_routine(compat.BytesIO(JSON))
        # Using a different generator to make the point that we can chain
        # user-provided code
        def event_yielder():
            for evt in events:
                yield evt
        results = list(routine(event_yielder(), *args, **kwargs))
        self.assertEqual(expected_results, results)

    def _assert_entry_point(self, expected_results, previous_routine, routine,
                            *args, **kwargs):
        self._assert_invalid_type(routine, *args, **kwargs)
        self._assert_bytes(expected_results, routine, *args, **kwargs)
        self._assert_str(expected_results, routine, *args, **kwargs)
        self._assert_unicode(expected_results, routine, *args, **kwargs)
        self._assert_file(expected_results, routine, *args, **kwargs)
        self._assert_async_file(expected_results, routine, *args, **kwargs)
        self._assert_async_types_coroutine(expected_results, routine, *args, **kwargs)
        if previous_routine:
            self._assert_events(expected_results, previous_routine, routine, *args, **kwargs)

    def test_rich_basic_parse(self):
        self._assert_entry_point(JSON_EVENTS, None, self.basic_parse)

    def test_rich_parse(self):
        self._assert_entry_point(JSON_PARSE_EVENTS, self.basic_parse, self.parse)

    def test_rich_items(self):
        self._assert_entry_point([JSON_OBJECT], self.parse, self.items, '')

    def test_rich_kvitems(self):
        self._assert_entry_point(JSON_KVITEMS, self.parse, self.kvitems, 'docs.item')

generate_backend_specific_tests(globals(), 'MainEntryPoints', '', MainEntryPoints)


try:
    from distutils import ccompiler
    from distutils import sysconfig
except ImportError:
    from setuptools._distutils import ccompiler
    from setuptools._distutils import sysconfig

import glob
import os
import platform
import shutil
import tempfile

from setuptools import setup, find_packages, Extension

def get_ijson_version():
    """Get version from code without fully importing it"""
    _globals = {}
    with open(os.path.join('ijson', 'version.py')) as f:
        code = f.read()
    exec(code, _globals)
    return _globals['__version__']

setupArgs = dict(
    name = 'ijson',
    version = get_ijson_version(),
    author = 'Rodrigo Tobar, Ivan Sagalaev',
    author_email = 'rtobar@icrar.org, maniac@softwaremaniacs.org',
    url = 'https://github.com/ICRAR/ijson',
    license = 'BSD',
    description = 'Iterative JSON parser with standard Python iterator interfaces',
    long_description = open('README.rst').read(),
    long_description_content_type = 'text/x-rst',

    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages = find_packages(exclude=['test']),
)

# Check if the yajl library + headers are present
# We don't use compiler.has_function because it leaves a lot of files behind
# without properly cleaning up
def yajl_present():

    compiler = ccompiler.new_compiler(verbose=1)
    sysconfig.customize_compiler(compiler) # CC, CFLAGS, LDFLAGS, etc

    fname = tempfile.mktemp(".c", "yajl_version")
    try:
        with open(fname, "wt") as f:
            f.write('''
            #include <yajl/yajl_version.h>
            int main(int args, char **argv)
            {
            #if YAJL_MAJOR != 2
                fail to compile
            #else
                yajl_version();
            #endif
                return 0;
            }
            ''')

        try:
            objs = compiler.compile([fname])
            compiler.link_shared_lib(objs, 'a', libraries=["yajl"])
            return True
        finally:
            os.remove(compiler.library_filename('a', lib_type='shared'))
            for obj in objs:
                os.remove(obj)

    except:
        return False
    finally:
        if os.path.exists(fname):
            os.remove(fname)


def patch_yajl_sources():
    """Make yajl sources ready for direct compilation against them"""
    # cp cextern/yajl -R $yajl_sources_copy
    # mkdir $yajl_sources_copy/yajl
    # cp $yajl_sources_copy/src/api/*.h $yajl_sources_copy/yajl
    patched_sources = os.path.join(tempfile.mkdtemp(), 'yajl')
    shutil.copytree(os.path.join('cextern', 'yajl'), patched_sources)
    headers_original = os.path.join(patched_sources, 'src', 'api')
    headers_copy = os.path.join(patched_sources, 'yajl')
    shutil.copytree(headers_original, headers_copy)
    return patched_sources


extra_sources = []
extra_include_dirs = []
libs = ['yajl']
embed_yajl = os.environ.get('IJSON_EMBED_YAJL', None) == '1'
if not embed_yajl:
    have_yajl = yajl_present()
else:
    yajl_sources = patch_yajl_sources()
    extra_sources = sorted(glob.glob(os.path.join(yajl_sources, 'src', '*.c')))
    extra_sources.remove(os.path.join(yajl_sources, 'src', 'yajl_version.c'))
    extra_include_dirs = [yajl_sources, os.path.join(yajl_sources, 'src')]
    libs = []
build_yajl_default = '1' if embed_yajl or have_yajl else '0'
build_yajl = os.environ.get('IJSON_BUILD_YAJL2C', build_yajl_default) == '1'
if build_yajl:
    yajl_ext = Extension('ijson.backends._yajl2',
                         language='c',
                         sources=sorted(glob.glob('ijson/backends/yajl2_c/*.c')) + extra_sources,
                         include_dirs=['ijson/backends/yajl2_c'] + extra_include_dirs,
                         libraries=libs,
                         depends=glob.glob('ijson/backends/yajl2_c/*.h'))
    setupArgs['ext_modules'] = [yajl_ext]

setup(**setupArgs)
